#    $Id: Data.pm,v 1.26 2004/02/05 22:47:34 kevinspicer Exp $
#    MailScanner-MRTG 
#    Copyright (C) 2002-3 Dale Lovelace <dlovelace@hotels.com>
#    With various bits shamelessly stolen from others
#    Restructuring and Substantial Modifications 
#    Copyright (C) 2003-4 Kevin Spicer <kevin@kevinspicer.co.uk>
#    with contributions from many other (see CHANGELOG for details)
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Please submit all bug reports, patches etc via the sourceforge project 
#   page at http://sourceforge.net/projects/mailscannermrtg
#
#

package MSMRTG::Data;
require Exporter;

@ISA=qw(Exporter);
@EXPORT=qw(GetData);

use FileHandle;

sub GetData()
{
	# Outer function that controls the data gathering process
	my ($mail, $spam1, $spam2, $virus1, $virus2, $resetvalue); # For holding temporary data

  if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("Beginning sub GetData") }

	# If the lastrun date isn't the same as today then wipe out the counters
	# for those that increment rather than being a fresh count each time
  if ($MSMRTG::Config::timedate{'mday'} != $MSMRTG::State::lastrun[0]) {
		# Reset to 0 if the time is before 00:05, otherwise to 'u'
		if ($MSMRTG::Config::timedate{'hour'} == 0 &&
						$MSMRTG::Config::timedate{'min'} <5) { $resetvalue = 0 }
		else {$resetvalue = 'u'}
	  foreach (qw(mail mailbytes virus spam))
	  {
      $MSMRTG::State::logdata{$_}[0]=$resetvalue;
      $MSMRTG::State::logdata{$_}[1]=$resetvalue;
	  }
  }


	
  # Collect the data
	
	#Data from the uptime command
  GetUptimeData();

	# Data from SNMP
	GetSNMPData();

	# Counting number of files
	GetFileCounts();

	# Counting processes
	GetPSData();

	# Counting filesystem usage
	GetFSUsage();

	# Get the current values for mail, spam and virus - before updating them
	$mail=$MSMRTG::State::logdata{'mail'}[0];
	$spam1=$MSMRTG::State::logdata{'spam'}[0];
	$spam2=$MSMRTG::State::logdata{'spam'}[1];
	$virus1=$MSMRTG::State::logdata{'virus'}[0];
	$virus2=$MSMRTG::State::logdata{'virus'}[1];
	# Notice we substract the number of actual viruses from $virus1
	# to try and reduce duplication
	if ($virus1 ne 'u' && $virus2 ne 'u') {$virus1-=$virus2};

	# Make sure that any of the log counters set to 'u' (undefined) are changed
	# to zero.
	foreach (qw(mailbytes mail spam virus))
	{
		if ($MSMRTG::State::logdata{$_}[0] eq 'u') { $MSMRTG::State::logdata{$_}[0]=0 }
		if ($MSMRTG::State::logdata{$_}[1] eq 'u') { $MSMRTG::State::logdata{$_}[1]=0 }
	}
	
  # Parse the log files 
  foreach (keys(%MSMRTG::State::logfiles)){

    ProcessFile($_);

  }

	# Do the virus and spam ratios
	GetRatios($mail, $spam1, $spam2, $virus1, $virus2);

  if ($MSMRTG::Debug::debug) {
    MSMRTG::Debug::Debug("Uptime", $MSMRTG::State::logdata{'uptime'}[0]);
    MSMRTG::Debug::Debug("CPU Total1", $MSMRTG::State::logdata{'cpu'}[0]);
    MSMRTG::Debug::Debug("CPU Total2", $MSMRTG::State::logdata{'cpu'}[1]);
    MSMRTG::Debug::Debug("Memory Usage Total1", $MSMRTG::State::logdata{'memory'}[0]);
    MSMRTG::Debug::Debug("Memory Usage Total2", $MSMRTG::State::logdata{'memory'}[1]);
    MSMRTG::Debug::Debug("IP Traffic Total1", $MSMRTG::State::logdata{'iptraffic'}[0]);
    MSMRTG::Debug::Debug("IP Traffic Total2", $MSMRTG::State::logdata{'iptraffic'}[1]);
    MSMRTG::Debug::Debug("Incoming Queue Total", $MSMRTG::State::logdata{'inqueue'}[0]);
    MSMRTG::Debug::Debug("Outgoing Queue Total", $MSMRTG::State::logdata{'outqueue'}[0]);
    MSMRTG::Debug::Debug("MailScanner Processes Total", $MSMRTG::State::logdata{'mailscanner'}[0]);
    MSMRTG::Debug::Debug("MTA processes Total", $MSMRTG::State::logdata{'mta'}[0]);
    MSMRTG::Debug::Debug("Root Filesystem Usage Total", $MSMRTG::State::logdata{'rootusage'}[0]);
    MSMRTG::Debug::Debug("Spool Filesystem Usage Total", $MSMRTG::State::logdata{'spoolusage'}[0]);
    MSMRTG::Debug::Debug("MailScanner Incoming Usage Total", $MSMRTG::State::logdata{'tmpfsusage'}[0]);

    MSMRTG::Debug::Debug("Mail Total", $MSMRTG::State::logdata{'mail'}[0]);
    MSMRTG::Debug::Debug("Viruses Total1", $MSMRTG::State::logdata{'virus'}[0]);
    MSMRTG::Debug::Debug("Viruses Total2", $MSMRTG::State::logdata{'virus'}[1]);
    MSMRTG::Debug::Debug("Mailbytes Total", $MSMRTG::State::logdata{'mailbytes'}[0]);
    MSMRTG::Debug::Debug("Spam Total1", $MSMRTG::State::logdata{'spam'}[0]);
    MSMRTG::Debug::Debug("Spam Total2", $MSMRTG::State::logdata{'spam'}[1]);
    MSMRTG::Debug::Debug("Virus Ratio1", $MSMRTG::State::logdata{'virusratio'}[0]);
    MSMRTG::Debug::Debug("Viruses Ratio2", $MSMRTG::State::logdata{'virusratio'}[1]);
    MSMRTG::Debug::Debug("Spam Ratio1", $MSMRTG::State::logdata{'spamratio'}[0]);
    MSMRTG::Debug::Debug("Spam Ratio2", $MSMRTG::State::logdata{'spamratio'}[1]);
  }
  
  if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("Leaving sub GetData") }
}

sub GetSNMPData {
  # This deals with retrieving data from SNMP
	# 
	# We actually use shell commands to collect this data, doubtless numerous
	# folks will point out that we could have used one of the available SNMP
	# modules for perl.  This is a deliberate descision not to - because
	# there are several SNMP modules and we don't want to cause confusion
	# We also allow people to turn of SNMP in the config file, so we don't want
	# to introduce a dependency that some people won't need
	# 
	my $command;
	my $snmpversion;

  if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("Entering sub GetSNMPData") }
	# Check we have been configured to use SNMP
	unless (defined($MSMRTG::Config::Config{'usesnmp'}))
	{
    if ($MSMRTG::Debug::debug) { 
			warn "Use SNMP missing in config file\n\tSkipping cpu, memory and iptraffic\n"; 
		}
		$MSMRTG::State::logdata{'cpu'} = ['u', 'u','u','u','u','u','u'];
		$MSMRTG::State::logdata{'memory'} = ['u','u','u'];
		$MSMRTG::State::logdata{'iptraffic'} = ['u','u','u'];
		return;
	}

	unless ($MSMRTG::Config::Config{'usesnmp'} =~ /yes/i)
	{
    if ($MSMRTG::Debug::debug) { 
			MSMRTG::Debug::Debug("SNMP turned off in config"); 
		}
		$MSMRTG::State::logdata{'cpu'} = ['u', 'u','u','u','u','u','u'];
		$MSMRTG::State::logdata{'memory'} = ['u','u','u'];
		$MSMRTG::State::logdata{'iptraffic'} = ['u','u','u'];
		return;
	}

	# Check a valid port number for snmp is being used
	unless ($MSMRTG::Config::Config{'snmpport'} =~ /^(\d+)$/)
	{
		warn "Non-numeric SNMP port specified in $MSMRTG::Config::configfile\n\tSkipping snmp functions\n\n";
		$MSMRTG::State::logdata{'cpu'} = ['u', 'u','u','u','u','u','u'];
		$MSMRTG::State::logdata{'memory'} = ['u','u','u'];
		$MSMRTG::State::logdata{'iptraffic'} = ['u','u','u'];
		return;
	}

  # Check a community string for snmp has been given
	unless (defined($MSMRTG::Config::Config{'snmpcommunity'}))
	{
		warn "No SNMP community specified in $MSMRTG::Config::configfile\n\tSkipping snmp functions\n\n";
		$MSMRTG::State::logdata{'cpu'} = ['u', 'u','u','u','u','u','u'];
		$MSMRTG::State::logdata{'memory'} = ['u','u','u'];
		$MSMRTG::State::logdata{'iptraffic'} = ['u','u','u'];
		return;
	}

  # Check the snmpwalk command is specified
	unless ($MSMRTG::Config::Config{"snmpwalkbinary"}) {
    warn "ERROR: Snmpwalk Binary not specified in $MSMRTG::Config::configfile\n\tSkipping snmp functions\n";
		$MSMRTG::State::logdata{'cpu'} = ['u', 'u','u','u','u','u','u'];
		$MSMRTG::State::logdata{'memory'} = ['u','u','u'];
		$MSMRTG::State::logdata{'iptraffic'} = ['u','u','u'];
		return;
  }

	unless (-x $MSMRTG::Config::Config{"snmpwalkbinary"}) {
		warn "ERROR: Snmpwalk Binary specified in $MSMRTG::Config::configfile is not executable or not present\n\tMaybe you need to install the snmp or snmp-utils packages\n\t\tSee the README.SNMP file in the docs\n\tSkipping snmp functions\n";
		$MSMRTG::State::logdata{'cpu'} = ['u', 'u','u','u','u','u','u'];
		$MSMRTG::State::logdata{'memory'} = ['u','u','u'];
		$MSMRTG::State::logdata{'iptraffic'} = ['u','u','u'];
		return;
	}

	$snmpversion=`$MSMRTG::Config::Config{'snmpwalkbinary'} -V 2>&1`;

  # Sort out the arguments to snmp commands	
	if ($snmpversion =~ /^net/i)
	{
		# net-snmp varient
		$command = "$MSMRTG::Config::Config{'snmpwalkbinary'} -v 2c -c $MSMRTG::Config::Config{'snmpcommunity'} localhost:$MSMRTG::Config::Config{'snmpport'} ";
	} else {
		# ucd-snmp varient
		$command = "$MSMRTG::Config::Config{'snmpwalkbinary'} -v 2c -p $MSMRTG::Config::Config{'snmpport'} localhost $MSMRTG::Config::Config{'snmpcommunity'} ";
	}

	getMemoryfromSNMP($command);
	getCpufromSNMP($command);
	getIptrafficfromSNMP($command);

  if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("Leaving sub GetSNMPData") }
}

sub getMemoryfromSNMP
{
	# Gets the memory statistics from SNMP
	# Takes a single argument, the SNMP command as formatted by GetSNMPData
	
	my $command=shift;
	my @snmpout;
	my $oid=".1.3.6.1.4.1.2021.4";
	my $memTotalReal=0;
	my $memAvailReal=0;
	my $memShared=0;
	my $memBuffer=0;
	my $memCached=0;

  if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("Entering sub GetMemoryfromSNMP") }


  @snmpout=`$command $oid`;

	foreach (@snmpout)
	{
		chomp;
    if ($MSMRTG::Debug::debug) { 
	    MSMRTG::Debug::Debug($_);
		}
		if (/.*memTotalReal.0 =.* (\d+)/) { $memTotalReal = $1 }
		elsif (/.*memAvailReal.0 =.* (\d+)/) { $memAvailReal = $1 }
		elsif (/.*memShared.0 =.* (\d+)/) { $memShared = $1 }
		elsif (/.*memBuffer.0 =.* (\d+)/) { $memBuffer = $1 }
		elsif (/.*memCached.0 =.* (\d+)/) { $memCached = $1 }
	}
	# (Not all the above appear in all SNMP versions - but thats okay becasue we
	# initialised them all as zero)
  if ($MSMRTG::Debug::debug) { 
		MSMRTG::Debug::Debug("Total Memory", $memTotalReal);
		MSMRTG::Debug::Debug("Available Memory", $memAvailReal); 
		MSMRTG::Debug::Debug("Shared Memory", $memShared);
		MSMRTG::Debug::Debug("Buffer Memory", $memBuffer); 
		MSMRTG::Debug::Debug("Cache Memory", $memCached); 
	}
	
  # Total used memory
  $MSMRTG::State::logdata{'memory'}[0] = $memTotalReal - $memAvailReal;
  # Total memory -actually- in use
  $MSMRTG::State::logdata{'memory'}[1] = $MSMRTG::State::logdata{'memory'}[0]-$memShared-$memBuffer-$memCached;
  # I want the output in MegaBytes!
  $MSMRTG::State::logdata{'memory'}[0] = sprintf("%.0f",($MSMRTG::State::logdata{'memory'}[0] / 1024));
  $MSMRTG::State::logdata{'memory'}[1] = sprintf("%.0f",($MSMRTG::State::logdata{'memory'}[1] / 1024));
	
	
  if ($MSMRTG::Debug::debug) { 
		MSMRTG::Debug::Debug("Memory Total1", $MSMRTG::State::logdata{'memory'}[0]);
		MSMRTG::Debug::Debug("Memory Total2", $MSMRTG::State::logdata{'memory'}[1]);
		MSMRTG::Debug::Debug("Leaving sub GetMemoryfromSNMP");
	}
}

sub getCpufromSNMP
{
	# Gets the CPU statistics from SNMP
	# Takes a single argument, the SNMP command as formatted by GetSNMPData
	
	my $command=shift;
	my @snmpout;
	my $oid=".1.3.6.1.4.1.2021.11"; # systemStats
	my $ssCpuRawIdle=0;
	my $ssCpuRawSystem=0;
	my $ssCpuRawUser=0;
	my $timeticks=0;
	my $idleCounter=32;
	my $userCounter=32;
	my $systemCounter=32;
	my ($idle, $system, $user, $totpc, $userpc);

  if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("Entering sub GetCpufromSNMP") }


  @snmpout=`$command $oid`;
	
	$oid=".1.3.6.1.2.1.1.3"; #System Uptime
  push @snmpout, `$command $oid`;

	foreach (@snmpout)
	{
	  chomp;
    if ($MSMRTG::Debug::debug) { 
	    MSMRTG::Debug::Debug($_);
		}
		if (/.*ssCpuRawUser.0 = Counter(\d+): (\d+)/) { 
			$ssCpuRawUser = $2;
			$userCounter = $1;
		}
		elsif (/.*ssCpuRawSystem.0 = Counter(\d+): (\d+)/) { 
			$ssCpuRawSystem = $2;
			$systemCounter = $1;
		}
		elsif (/.*ssCpuRawIdle.0 = Counter(\d+): (\d+)/) { 
			$ssCpuRawIdle = $2;
			$idleCounter = $1;
		}
		elsif (/.*sysUpTime.0 = Timeticks: \((\d+)\)/) { 
			$timeticks = $1;
		}
	}

  if ($MSMRTG::Debug::debug) { 
		MSMRTG::Debug::Debug("Idle Counter", $ssCpuRawIdle);
		MSMRTG::Debug::Debug("Idle Counter Size", $idleCounter);
		MSMRTG::Debug::Debug("System Counter", $ssCpuRawSystem); 
		MSMRTG::Debug::Debug("System Counter Size", $systemCounter); 
		MSMRTG::Debug::Debug("User Counter", $ssCpuRawUser);
		MSMRTG::Debug::Debug("User Counter Size", $userCounter);
		MSMRTG::Debug::Debug("Uptime (timeticks)", $timeticks);
	}

	# If we don't have a previous value then just set current value as previous
	# and return.
	if (
		(!defined ($MSMRTG::State::logdata{'cpu'}[3]) || $MSMRTG::State::logdata{'cpu'}[3] eq 'u') ||
		(!defined ($MSMRTG::State::logdata{'cpu'}[4]) || $MSMRTG::State::logdata{'cpu'}[4] eq 'u') ||
		(!defined ($MSMRTG::State::logdata{'cpu'}[5]) || $MSMRTG::State::logdata{'cpu'}[5] eq 'u') ||
		(!defined ($MSMRTG::State::logdata{'cpu'}[6]) || $MSMRTG::State::logdata{'cpu'}[5] eq 'u') 
		)
	{
		$MSMRTG::State::logdata{'cpu'}[0]='u';
		$MSMRTG::State::logdata{'cpu'}[1]='u';
		$MSMRTG::State::logdata{'cpu'}[3]=$ssCpuRawIdle;
		$MSMRTG::State::logdata{'cpu'}[4]=$ssCpuRawUser;
		$MSMRTG::State::logdata{'cpu'}[5]=$ssCpuRawSystem;
		$MSMRTG::State::logdata{'cpu'}[6]=$timeticks;
    if ($MSMRTG::Debug::debug) { 
	    MSMRTG::Debug::Debug("Cpu counters not initialised - no CPU data on this run");
		}
		return;
	}

	
	# Look out - if the previous run was too long ago (say more than 7
	# minutes) don't register any data this time
	# Similarly we can tell if the system has been restarted from the 
	# timeticks
	if (
		$MSMRTG::Config::timedate{'utime'}-$MSMRTG::State::lastrun[1] > 420 ||
		$MSMRTG::State::logdata{'cpu'}[6] > $timeticks
		)
	{
		$MSMRTG::State::logdata{'cpu'}[0]='u';
		$MSMRTG::State::logdata{'cpu'}[1]='u';
		$MSMRTG::State::logdata{'cpu'}[3]=$ssCpuRawIdle;
		$MSMRTG::State::logdata{'cpu'}[4]=$ssCpuRawUser;
		$MSMRTG::State::logdata{'cpu'}[5]=$ssCpuRawSystem;
		$MSMRTG::State::logdata{'cpu'}[6]=$timeticks;
    if ($MSMRTG::Debug::debug) { 
	    MSMRTG::Debug::Debug("Cpu counter data too old, or system restarted  - no CPU data on this run");
		}
		return;
	}

	# Maybe the counters this time are lower than last time we checked?
	# This will be because they wrapped around.
	# We can correct this because we caught the size of the counters
	
	if ($MSMRTG::State::logdata{'cpu'}[3] > $ssCpuRawIdle) {
		$MSMRTG::State::logdata{'cpu'}[3] -= 2**$idleCounter;
	}
	if ($MSMRTG::State::logdata{'cpu'}[4] > $ssCpuRawUser) {
		$MSMRTG::State::logdata{'cpu'}[4] -= 2**$userCounter;
	}
	if ($MSMRTG::State::logdata{'cpu'}[5] > $ssCpuRawSystem) {
		$MSMRTG::State::logdata{'cpu'}[5] -= 2**$systemCounter;
	}

	$idle = $ssCpuRawIdle - $MSMRTG::State::logdata{'cpu'}[3];
	$user = $ssCpuRawUser - $MSMRTG::State::logdata{'cpu'}[4];
	$system = $ssCpuRawSystem - $MSMRTG::State::logdata{'cpu'}[5];
	$MSMRTG::State::logdata{'cpu'}[3] = $ssCpuRawIdle;
	$MSMRTG::State::logdata{'cpu'}[4] = $ssCpuRawUser;
	$MSMRTG::State::logdata{'cpu'}[5] = $ssCpuRawSystem;
	$MSMRTG::State::logdata{'cpu'}[6] = $timeticks;
	
	if ($MSMRTG::Debug::debug) { 
		MSMRTG::Debug::Debug("Idle", $idle);
		MSMRTG::Debug::Debug("User", $user);
		MSMRTG::Debug::Debug("System", $system);
	}

	$totpc=$userpc=0;
	if (($user + $system) > 0){
	  $totpc = (($user + $system) * 100) / ($user + $system + $idle);
	}
	if ($user > 0) {
	  $userpc = ($user * 100) / ($user + $system + $idle);
	}
  
  # We output the total of User + System CPU Utilization
	# use snprintf to round the number
	$MSMRTG::State::logdata{'cpu'}[0]=sprintf("%.0f",$totpc);
	# Second line on graph shows user process CPU Utilization
	$MSMRTG::State::logdata{'cpu'}[1]=sprintf("%.0f",$userpc);
	
	
  if ($MSMRTG::Debug::debug) { 
		MSMRTG::Debug::Debug("Cpu Total1", $MSMRTG::State::logdata{'cpu'}[0]);
		MSMRTG::Debug::Debug("Cpu Total2", $MSMRTG::State::logdata{'cpu'}[1]);
		MSMRTG::Debug::Debug("Leaving sub GetCpufromSNMP");
	}
}

sub getIptrafficfromSNMP
{
	# Gets the Network statistics from SNMP
	# Takes a single argument, the SNMP command as formatted by GetSNMPData
	
	my $command=shift;
	my @snmpout;
	my %nwdata;
	my %nwindex;
	my $interfaces;
	my $oid=".1.3.6.1.2.1.2.2.1.2"; # interface descriptions
	my $okay=1;
	my $count=0;
	my $temp;
	my $element;
	my $timeticks=0;
	my $inbytes=0;
	my $outbytes=0;

  if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("Entering sub GetIptrafficfromSNMP") }

	# Check that the interfaces to monitor have been specified
	if (!defined ($MSMRTG::Config::Config{'interfacestomonitor'}))
	{
		warn "ERROR: Interfaces to Monitor not specified in $MSMRTG::Config::configfile\n\tSkipping iptraffic\n" ;
		$MSMRTG::State::logdata{'iptraffic'}=['u', 'u', 'u'];
		return;
	}
	$interfaces= "(" . (join ')|(', (split /\s*,\s*/, $MSMRTG::Config::Config{'interfacestomonitor'})) . ")";

  @snmpout=`$command $oid`;
	
	$oid=".1.3.6.1.2.1.2.2.1.10"; # incoming Bytes
  push @snmpout, `$command $oid`;
	$oid=".1.3.6.1.2.1.2.2.1.16"; # outgoing Bytes
  push @snmpout, `$command $oid`;
	$oid=".1.3.6.1.2.1.1.3"; #System Uptime
  push @snmpout, `$command $oid`;

	foreach (@snmpout)
	{
	  chomp;
    if ($MSMRTG::Debug::debug) { 
	    MSMRTG::Debug::Debug($_);
		}
		if (/.*ifDescr.(\d+) = (?:STRING: )?(\w+)/) { 
			$nwdata{$1}{'name'}=$2;
		}
		elsif (/.*ifOutOctets.(\d+) = Counter(\d+): (\d+)/) { 
			$nwdata{$1}{'outcounter'}=$2;
			$nwdata{$1}{'outbytes'}=$3;
		}
		elsif (/.*ifInOctets.(\d+) = Counter(\d+): (\d+)/) { 
			$nwdata{$1}{'incounter'}=$2;
			$nwdata{$1}{'inbytes'}=$3;
		}
		elsif (/.*sysUpTime.0 = Timeticks: \((\d+)\)/) { 
			$timeticks = $1;
		}
	}

	# Discard any interfaces which don't match the ones we want.
	foreach $int (keys(%nwdata))
	{
		if ($nwdata{$int}{'name'} !~ /$interfaces/oi)
		  { delete $nwdata{$int} }
	}

  if ($MSMRTG::Debug::debug) { 
		foreach $int (keys(%nwdata))
		{
		  MSMRTG::Debug::Debug("Interface" , $nwdata{$int}{'name'});
		  MSMRTG::Debug::Debug("Incoming traffic" , $nwdata{$int}{'inbytes'} . " (Counter" . $nwdata{$int}{'incounter'} . ")");
		  MSMRTG::Debug::Debug("Outgoing traffic" , $nwdata{$int}{'outbytes'} . " (Counter" . $nwdata{$int}{'outcounter'} . ")");
		}
	}

	foreach (keys(%nwdata))
	{
		$nwindex{$nwdata{$_}{'name'}}=$_;
	}

	# Grab the saved data from the config file
	#
  $temp=$MSMRTG::State::logdata{'iptraffic'};	
	splice(@$temp,0,3);
	$oldtimeticks=shift(@$temp);
	if (! defined($$temp[0])) { 
		$okay=0;
	}

	while (defined($$temp[0]))
	{
		$element=shift(@$temp);
		if (! exists ($nwindex{$element}) )
		{
			$okay=0;
			last;
		}
		if (! defined $$temp[0])
		{
			$okay=0;
			last;
		}
		$nwdata{$nwindex{$element}}{'oldinbytes'}=shift(@$temp);
		if (! defined $$temp[0])
		{
			$okay=0;
			last;
		}
		$nwdata{$nwindex{$element}}{'oldoutbytes'}=shift(@$temp);
		if ($nwdata{$nwindex{$element}}{'oldinbytes'} !~ /^\d+$/ ||
		    $nwdata{$nwindex{$element}}{'oldinbytes'} !~ /^\d+$/)
		{
			$okay=0;
			last;
		}
		$count++;
	}

	# Check we got the correct number of interfaces and values where required
	# If not save the current values and return
	if (! $okay || $count != scalar(keys(%nwdata))) {
		warn "ERROR: iptraffic counters not fully initialised\n\tNo iptraffic data on this run\nThis is normal on the first run, or when the previous run was more than 5 minutes ago (e.g. when rebooting).  If these errors persist over multiple runs please check that snmpd is running on your system.  If you don't want to run snmp please ensure that you have disabled it in mailscanner-mrtg.conf.  For more hints about snmp see README.SNMP in the docs directory\n";
		$MSMRTG::State::logdata{'iptraffic'}=['u', 'u', 'u', $timeticks];
		$count=4;
		foreach (keys(%nwdata))
		{
			$MSMRTG::State::logdata{'iptraffic'}[$count++]= $nwdata{$_}{'name'};
			$MSMRTG::State::logdata{'iptraffic'}[$count++]= $nwdata{$_}{'inbytes'};
			$MSMRTG::State::logdata{'iptraffic'}[$count++]= $nwdata{$_}{'outbytes'};
		}
		return;
	}	

	# Look out - if the previous run was too long ago (say more than 7
	# minutes) don't register any data this time
	# Similarly we can tell if the system has been restarted from the 
	# timeticks
	if (
		$MSMRTG::Config::timedate{'utime'}-$MSMRTG::State::lastrun[1] > 420 ||
		$oldtimeticks > $timeticks
		)
	{
		$MSMRTG::State::logdata{'iptraffic'}=['u', 'u', 'u', $timeticks];
		$count=4;
		foreach (keys(%nwdata))
		{
			$MSMRTG::State::logdata{'iptraffic'}[$count++]= $nwdata{$_}{'name'};
			$MSMRTG::State::logdata{'iptraffic'}[$count++]= $nwdata{$_}{'inbytes'};
			$MSMRTG::State::logdata{'iptraffic'}[$count++]= $nwdata{$_}{'outbytes'};
		}
    if ($MSMRTG::Debug::debug) { 
	    MSMRTG::Debug::Debug("IP traffic data too old, or system restarted  - noiptraffic data on this run");
		}
		return;
	}

  if ($MSMRTG::Debug::debug) { 
		foreach $int (keys(%nwdata))
		{
		  MSMRTG::Debug::Debug("Interface" , $nwdata{$int}{'name'});
		  MSMRTG::Debug::Debug("Previous Incoming traffic" , $nwdata{$int}{'oldinbytes'} . " (Counter" . $nwdata{$int}{'incounter'} . ")");
		  MSMRTG::Debug::Debug("Previous Outgoing traffic" , $nwdata{$int}{'oldoutbytes'} . " (Counter" . $nwdata{$int}{'outcounter'} . ")");
		}
	}
	
	foreach (keys(%nwdata))
	{
		# Maybe the counters this time are lower than last time we checked?
		# This will be because they wrapped around.
		# We can correct this because we caught the size of the counters
		if ($nwdata{$_}{'oldinbytes'} > $nwdata{$_}{'inbytes'}) {
			$nwdata{$_}{'oldinbytes'} -= 2**$nwdata{$_}{'incounter'}
		}
		if ($nwdata{$_}{'oldoutbytes'} > $nwdata{$_}{'outbytes'}) {
			$nwdata{$_}{'oldoutbytes'} -= 2**$nwdata{$_}{'outcounter'}
		}
		# And (finally!) calculate the bytes totals
		$inbytes += (($nwdata{$_}{'inbytes'} - $nwdata{$_}{'oldinbytes'}));
		$outbytes += (($nwdata{$_}{'outbytes'} - $nwdata{$_}{'oldoutbytes'}));
	}

	if ($MSMRTG::Debug::debug) { 
		MSMRTG::Debug::Debug("Incoming Bytes", $inbytes);
		MSMRTG::Debug::Debug("Outgoing Bytes", $outbytes);
	}

	# We want to turn this into bits per second
	$inbytes /= 0.125*
		($MSMRTG::Config::timedate{'utime'}-$MSMRTG::State::lastrun[1]); 
	$outbytes /= 0.125*
		($MSMRTG::Config::timedate{'utime'}-$MSMRTG::State::lastrun[1]); 

	$MSMRTG::State::logdata{'iptraffic'}=();
  # We output the total of User + System CPU Utilization
	# use snprintf to round the number
	$MSMRTG::State::logdata{'iptraffic'}[0]=sprintf("%.0f",$inbytes);
	# Second line on graph shows user process CPU Utilization
	$MSMRTG::State::logdata{'iptraffic'}[1]=sprintf("%.0f",$outbytes);
	
	$MSMRTG::State::logdata{'iptraffic'}[2] = 'u';
	$MSMRTG::State::logdata{'iptraffic'}[3] = $timeticks;

	$count=4;
	foreach (keys(%nwdata))
	{
		$MSMRTG::State::logdata{'iptraffic'}[$count++]= $nwdata{$_}{'name'};
		$MSMRTG::State::logdata{'iptraffic'}[$count++]= $nwdata{$_}{'inbytes'};
		$MSMRTG::State::logdata{'iptraffic'}[$count++]= $nwdata{$_}{'outbytes'};
	}
	
  if ($MSMRTG::Debug::debug) { 
		MSMRTG::Debug::Debug("IPtraffic Total1", $MSMRTG::State::logdata{'iptraffic'}[0]);
		MSMRTG::Debug::Debug("IPtraffic Total2", $MSMRTG::State::logdata{'iptraffic'}[1]);
		MSMRTG::Debug::Debug("Leaving sub GetIptrafficfromSNMP");
	}
}


sub GetUptimeData {
	# Finds average load average for last five minutes
	# And the uptime of the machine
  my($uptime);

  if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("Beginning sub GetUptimeData") }

  unless ($MSMRTG::Config::Config{"uptimecommand"}) {
    warn "ERROR: UpTime Command not specified in $MSMRTG::Config::configfile\n\tSkipping uptime data";
 	  $MSMRTG::State::logdata{'loadavg'}[0]='u';
 	  $MSMRTG::State::logdata{'loadavg'}[1]='u';
    $MSMRTG::State::logdata{'uptime'}[0]='u';
    $MSMRTG::State::logdata{'uptime'}[1]='u';
		return;
  }

	unless (-x $MSMRTG::Config::Config{"uptimecommand"}) {
		warn "ERROR: Uptime Binary specified in $MSMRTG::Config::configfile is not executable or not present\n\tSkipping uptime data\n";
 	  $MSMRTG::State::logdata{'loadavg'}[0]='u';
 	  $MSMRTG::State::logdata{'loadavg'}[1]='u';
    $MSMRTG::State::logdata{'uptime'}[0]='u';
    $MSMRTG::State::logdata{'uptime'}[1]='u';
		return;
	}

	#Run the uptime command
	$uptime=`$MSMRTG::Config::Config{"uptimecommand"}`;

  if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("line", $uptime) }

	# Get the Load average
	if ($uptime=~/load averages?:\s+\d+\.\d+,\s+(\d+\.\d+),\s+\d+\.\d+/) {
	  # Use sprintf to round up the decimal place
	  $MSMRTG::State::logdata{'loadavg'}[0]=sprintf("%.0f", ($1 * 100));
	  $MSMRTG::State::logdata{'loadavg'}[1]='u';
	} else {
    warn "ERROR: Load Average can not be computed\n\tSkipping loadavg\n";
	  $MSMRTG::State::logdata{'loadavg'}[0]='u';
	  $MSMRTG::State::logdata{'loadavg'}[1]='u';
	}

	# And the system uptime
  # Thanks to Kris Stumpner for this new and improved GetUpTime!
  # If your box is up less than 1 day the first regex will hit.
  # The second hits for uptime > 1 day!  [& <=1hr]
	# Now slightly revised second re now matches on Solaris
  if ($uptime =~ /^\s*\d+:\d+.+\s+up\s+(\d+:\d+.*?)/) {
    $MSMRTG::State::logdata{'uptime'}[0]=$1;
  } elsif ($uptime =~ /^\s*\d+:\d+.+\s+up\s+(\d+)\s+([\w()]+)/) {
    $MSMRTG::State::logdata{'uptime'}[0]=$1 . " " . $2;
  } else {
    warn "ERROR: UpTime can not be computed\n\tSkipping uptime\n";
    $MSMRTG::State::logdata{'uptime'}[0]='u';
  }

  $MSMRTG::State::logdata{'uptime'}[1]='u';


  if ($MSMRTG::Debug::debug) {
    MSMRTG::Debug::Debug("LoadAvg", $MSMRTG::State::logdata{'loadavg'}[0]);
    MSMRTG::Debug::Debug("Uptime", $MSMRTG::State::logdata{'uptime'}[0]);
    MSMRTG::Debug::Debug("Leaving sub GetUptimeData");
  }
}

sub countFiles {
	# Counts and returns number of files in directory given as first argument
	local $Total=0;

  if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("Beginning sub countFiles") }
  use File::Find;
  if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug ("use File::Find", "True") }	
	# This now recurses, counting only files - this is needed to support
	# Postfix and Exim
	File::Find::find({wanted => \&wanted}, $_[0]);

  if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("Leaving sub countFiles") }

	return ($Total);

}

sub GetFileCounts {
	my @tocount;
	my $datatype;

  if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("Beginning sub GetFileCounts") }

	@tocount = (
		[ "Incoming Queue Dir", "incomingqueuedir", "inqueue" ],
		[ "Outgoing Queue Dir", "outgoingqueuedir", "outqueue" ],
		[ "Quarantine Directory", "quarantinedirectory", "quarantine" ]
	);

	foreach (@tocount) {
  
		$datatype=$$_[2];

		unless ($MSMRTG::Config::Config{$$_[1]}) {
    	warn "ERROR: $$_[0] not specified in $MSMRTG::Config::configfile\n\tSkipping $$_[2]\n";
 	  	$MSMRTG::State::logdata{$datatype}[0]='u';
 	  	$MSMRTG::State::logdata{$datatype}[1]='u';
			next;
		}
		
 	 	$MSMRTG::State::logdata{$datatype}[0]=countFiles($MSMRTG::Config::Config{$$_[1]});
 	 	$MSMRTG::State::logdata{$datatype}[1]='u';

    if ($MSMRTG::Debug::debug) {
      MSMRTG::Debug::Debug("Files in $$_[0]", $MSMRTG::State::logdata{$datatype}[0]);
		}
	}
  
	if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("Leaving sub GetFileCounts") }

}

sub GetPSData {
	# Count MTA and MailScanner processes
	my $mtapsre; # regular expression for matching MTA processes
	my @ps;

	$MSMRTG::State::logdata{'mta'}[0]=0;
	$MSMRTG::State::logdata{'mta'}[1]='u';
	$MSMRTG::State::logdata{'mailscanner'}[0]=0;
	$MSMRTG::State::logdata{'mailscanner'}[1]='u';

  if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("Beginning sub MTA") }

	# Decide which MTA we are using
	if (lc($MSMRTG::Config::Config{"mta"}) eq "postfix") {
		$mtapsre="\\bmaster\\b";
	} elsif (lc($MSMRTG::Config::Config{"mta"}) eq "exim") {
		$mtapsre="\\bexim\\b";
	# Sendmail is the default if no MTA specified
	} elsif (lc($MSMRTG::Config::Config{"mta"}) eq "sendmail" || ! $MSMRTG::Config::Config{"mta"}) {
		$mtapsre="\\b(?:sendmail(?:-in)?)|(?:sm-mta)\\b";
	} else {
		warn "ERROR: " . $MSMRTG::Config::Config{"mta"} . " is not a supported MTA\n\tSkipping mta\n";
	  $MSMRTG::State::logdata{'mta'}[0]='u';
  	$MSMRTG::State::logdata{'mta'}[1]='u';
		$mtapsre='^+++THIS WILL NEVER MATCH+++$';
	}

	# Adjust the ps command for Sun Solaris
	if ($^O eq "solaris"){
  	@ps = `ps -eo args`;
	} elsif ($^O eq "freebsd"){
  	@ps = `ps axo command`;
	} else {
  	@ps = `ps axo comm`;
	}

  foreach $p (@ps) { 
    if ($p =~ /$mtapsre/o) { 
      $MSMRTG::State::logdata{'mta'}[0] ++; 
    }
    elsif ($p =~ /MailScanner/) { $MSMRTG::State::logdata{'mailscanner'}[0] ++ }
  }

  # NOTE TO SELF - THIS DOESN'T BELOW IN A DATA COLLECTION ROUTINE
	# we probably should have a seperate section for actions based on thresholds
	
	# restart MailScanner if number of MS processes is less than the threshold
	# specified in the config file
  if ($MSMRTG::Config::Config{"restartthreshhold"}) {
					
  if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("Checking sufficient MailScanner processes are running") }

    if ($MSMRTG::State::logdata{'mailscanner'}[0] < $MSMRTG::Config::Config{"restartthreshhold"}) {
			# Lets throw in a warn statement so the root user gets mail
			warn "Number of MailScanner processes below threshold, restarting MailScanner\n";
      if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("Restarting MailScanner") }
      `$MSMRTG::Config::Config{"restartmailscanner"}`;
    }
  }
  elsif ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("MailScanner restart checks disabled (Restart Threshold = 0)") }

  if ($MSMRTG::Debug::debug) {
    MSMRTG::Debug::Debug("MTA processes", $MSMRTG::State::logdata{'mta'}[0]);
    MSMRTG::Debug::Debug("MailScanner processes", $MSMRTG::State::logdata{'mailscanner'}[0]);
    MSMRTG::Debug::Debug("Leaving sub GetPSData");
  }
}

sub GetFSUsage {
	# Space used by the filesystems
	# It makes little sense to return a figure that doesn't correspond with
	# a mountpoint, so these now return 'u'
	my @ps;
	my $tmpfsdir;
	my $spooldir;
	my $use;
	my $mointpoint;

	$MSMRTG::State::logdata{'rootusage'}[0]='u';
	$MSMRTG::State::logdata{'rootusage'}[1]='u';
	$MSMRTG::State::logdata{'spoolusage'}[0]='u';
	$MSMRTG::State::logdata{'spoolusage'}[1]='u';
	$MSMRTG::State::logdata{'tmpfsusage'}[0]='u';
	$MSMRTG::State::logdata{'tmpfsusage'}[1]='u';

  if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("Beginning sub GetFSUsage") }

  if ($MSMRTG::Config::Config{'mailscannerworkdirectory'}){
    $tmpfsdir=$MSMRTG::Config::Config{'mailscannerworkdirectory'};
	} else {
		warn "MailScanner Work Directory not specified in $MSMRTG::Config::configfile\n\tSkipping tmpfsusage\n";
		$tmpfsdir="+++THIS WILL NOT MATCH+++";
	}

	if ($MSMRTG::Config::Config{'spooldirectory'}){
    $spooldir=$MSMRTG::Config::Config{'spooldirectory'};
	} else {
		warn "Spool Directory not specified in $MSMRTG::Config::configfile\n\tSkipping spoolusage\n";
		$spooldir="+++THIS WILL NOT MATCH+++";
	}
  @df = `df -k`;

	foreach (@df){
		if ( /^Filesystem/ ) { next }

		( undef, undef, undef, undef, $use, $mountpoint ) = split(/\s+/);
		
		if (! defined($use)) { next }
		
		if ($MSMRTG::Debug::debug) { MSMRTG::Debug::Debug("Space used in $mointpoint", $use) }

		chop $use;

		if ($mountpoint eq "/") { $MSMRTG::State::logdata{'rootusage'}[0]=$use }
		if ($mountpoint eq $spooldir) { $MSMRTG::State::logdata{'spoolusage'}[0]=$use}
		if ($mountpoint eq $tmpfsdir) { $MSMRTG::State::logdata{'tmpfsusage'}[0]=$use}
	}

	if ($MSMRTG::State::logdata{'spoolusage'}[0] eq 'u' && $spooldir !~ /^\+\+\+/) {
		warn "Unable to find a mountpoint for $spooldir.  Please set Spool Directory in mailscanner-mrtg.conf to a valid mountpoint\n";
	}
	if ($MSMRTG::State::logdata{'tmpfsusage'}[0] eq 'u' && $tmpfsdir !~ /^\+\+\+/) {
		warn "Unable to find a mountpoint for $tmpfsdir.  Please set MailScanner Work Directory in mailscanner-mrtg.conf to a valid mountpoint\n";
	}

	if ($MSMRTG::Debug::debug) { 
		if ($MSMRTG::State::logdata{'rootusage'}[0] ne 'u') {
		MSMRTG::Debug::Debug("rootusage", $MSMRTG::State::logdata{'rootusage'}[0] . "%"); 
		} else {
		MSMRTG::Debug::Debug("rootusage could not be determined"); 
		}
		if ($MSMRTG::State::logdata{'spoolusage'}[0] ne 'u') {
		MSMRTG::Debug::Debug("spoolusage", $MSMRTG::State::logdata{'spoolusage'}[0] . "%"); 
		} else {
		MSMRTG::Debug::Debug("spoolusage could not be determined"); 
		}
		if ($MSMRTG::State::logdata{'tmpfsusage'}[0] ne 'u') {
		MSMRTG::Debug::Debug("tmpfsusage", $MSMRTG::State::logdata{'tmpfsusage'}[0] . "%"); 
		} else {
		MSMRTG::Debug::Debug("tmpfsusage could not be determined"); 
		}
  	MSMRTG::Debug::Debug("Leaving sub GetFSUsage"); 
	}
}

sub wanted {
	# Called by file::find to count files
   my ($dev,$ino,$mode,$nlink,$uid,$gid);

   (($dev,$ino,$mode,$nlink,$uid,$gid) = lstat($_)) &&
   -f _ && $Total++;
	
}

#
# This function handles processing the log files, for the calls that require
# it.  The idea of combining these operations into a single function is that 
# it is then only necessary to make one pass through each log file.

sub ProcessFile{
  # This function does the actual processing of the log files 
  # It takes one arguments, the log file name

  my $pos=0;
  my $spam=0;
  my $blocked=0;
  my $viruses=0;
  my $problems=0;
	my $mailbytes=0;
  my $logfile=$_[0];
  my $mbfromms=0;
  my $mbfrommta=0;
	my $mtainlog=0;
	my $mtaoutlog=0;
	my $mtarejlog=0;
	my $mslog=0;
	my $mtareciplog=0;
	my $mtarecipre; # regular expression for counting recipients from MTA log
	my $mtamessre; # regular expression for counting messages from MTA log
	my $mtaspamre; # regular expression for counting spam rejects by MTA
	my $mtabytesre; # regular expression for counting mailbytes from MTA log
	my $mtanamere; # regular expression for matching MTA process name
	my $recipsbyline=0; # Whether recipients are counted by matching whole lines
	           # (normally - sendmail & postfix) they are pulled from a numeric 
						 # field in the line.
  
  if ($MSMRTG::Debug::debug) {
    MSMRTG::Debug::Debug("Beginning sub ProcessFile");
  }

	# Check which logs we are processing
	if ($logfile =~ /^\Q$MSMRTG::Config::Config{'mtaincominglog'}\E$/) {$mtainlog=1;}
	if ($logfile =~ /^\Q$MSMRTG::Config::Config{'mtaoutgoinglog'}\E$/) {$mtaoutlog=1;}
	if ($logfile =~ /^\Q$MSMRTG::Config::Config{'mtarejectlog'}\E$/) {$mtarejlog=1;}
	if ($logfile =~ /^\Q$MSMRTG::Config::Config{'mailscannerlog'}\E$/) {$mslog=1;}

	# Check which MTA we are using
	# Postfix
	if (lc($MSMRTG::Config::Config{"mta"}) eq "postfix") {
		$mtanamere="postfix\\/(?:n)?qmgr\\[\\d+\\]";
	 	$mtarecipre="nrcpt=(\\d+)";
	 	$mtamessre="nrcpt=[1-9]\\d*";  
		$mtaspamre="reject:.+55\\d.+";
		$mtabytesre="size=(\\d+)";
		# Postfix counts recipients in incoming mail
		$mtareciplog=$mtainlog;
		# And records the number of recipients in a single log line
		$recipsbyline=0;
	
  # Exim
	} elsif (lc($MSMRTG::Config::Config{"mta"}) eq "exim") {
		$mtanamere="."; # exim logs to its own log so doesn't include its name
	 	$mtarecipre="^(?:\\S+\\s+){3}[=-]>";
	 	$mtamessre="^(?:\\S+\\s+){3}<=";   
		$mtaspamre="rejected\\s+RCPT";
		$mtabytesre="<=\\s+.*\\s+P=\\w+\\s+S=(\\d+)";
		# Exim counts recipients in outgoing mail
		$mtareciplog=$mtaoutlog;
		# And records a seperate line in outgoing log for each recipient
		$recipsbyline=1;

	# Sendmail is the default if no MTA specified
	} elsif (lc($MSMRTG::Config::Config{"mta"}) eq "sendmail" || ! $MSMRTG::Config::Config{"mta"}) {
		$mtanamere="(?:sendmail(?:-in)?)|(?:sm-mta)\\[\\d+\\]";
		#$mtanamere="sendmail(?:-in)?\\[\\d+\\]";
	 	$mtarecipre= "nrcpts=(\\d+),";
	 	$mtamessre="nrcpts=[1-9]\\d*,";  
		$mtaspamre= "reject=55\\d";
		#$mtabytesre=".+sendmail(?:-in)?\\[\\d+\\]:\\s+\\w+:\\s+from=\\S+,\\s+size=(\\d+),.+";
		$mtabytesre="from=\\S+,\\s+size=(\\d+),";
		# Sendmail counts recipients in incoming mail
		$mtareciplog=$mtainlog;
		# And records the number of recipients in a single log line
		$recipsbyline=0;
	
	# Something else (unsupported)
	} else {
		warn "ERROR: " . $MSMRTG::Config::Config{"mta"} . " is not a supported MTA\n\tSkipping processing of $logfile\n";
		return;
	}

  # Decide where mailbytes are coming from (this should now be mta or
	# mailscanner but we'll preserve sendmail as a synonym for mta
  if (lc($MSMRTG::Config::Config{"calculatemailbytes"}) eq "mta" ||
      lc($MSMRTG::Config::Config{"calculatemailbytes"}) eq "sendmail") {
    $mbfrommta=1;
		
  } elsif (lc($MSMRTG::Config::Config{"calculatemailbytes"}) eq "mailscanner") {
    $mbfromms=1;

  } else {
    warn "ERROR: Calculate Mail Bytes value not specified in $MSMRTG::Config::configfile\n\tSkipping processing of $logfile\n";
		return;
  }

  unless (open(LOG, $logfile))
    {
			warn("Cannot access log file $logfile, $!\n\tSkipping processing of $logfile\n");
			return;
		}

  # Check that the currently recorded position isn't beyond the end of
  # the file (indicating the logs have rotated).
  seek(LOG,0,2);
  $pos = tell(LOG);
  if ($pos < $MSMRTG::State::logfiles{$logfile}) {
    $MSMRTG::State::logfiles{$logfile}=0;
  }
  seek(LOG, $MSMRTG::State::logfiles{$logfile},0);
    
  while(<LOG>) {
    chomp;
    # (The second argument indicates we should check we have todays logs)
		# This RE now accounts for the date format in exim's logs too
    if ($MSMRTG::State::firstrun && 
				! /^($MSMRTG::Config::timedate{'month'} \s+ 
						$MSMRTG::Config::timedate{'mday'}) | # syslog dd mmm hh:mm:ss
				($MSMRTG::Config::timedate{'year'} - $MSMRTG::Config::timedate{'mon'} 
				    - $MSMRTG::Config::timedate{'mday'} \s*
						\d+:\d+:\d+\s+ )  # exim yyyy-mm-dd hh:mm:ss
										 /x) { next; }
		# Lines logged by MailScanner processes
    if ($mslog && /MailScanner\[\d+\]/i) { 
      # The Virus logs
      if (/Found (\d+) (viruses)/i) { $viruses += $1 ; }
      elsif (/Found (\d+) (problems)/i)  { $problems += $1; }
    
      # The Spam logs
      elsif (/Spam Checks: Found (\d+) spam messages/i) { $spam += $1 }

      # Mailbytes if coming from mailscanner
      elsif ($mbfromms && /.+MailScanner\[\d+\]:\s+\w+(?:\s+\w+)?\s*:\s+\w+\s+\d+\s+\w+,\s+(\d+)\s+bytes.*/i) 
        { $mailbytes += $1; }

		# Lines logged by MTA processes
    } elsif (/$mtanamere/oi) {
      
			# The Mail logs (count recipients)
      if ($mtareciplog && /$mtarecipre/oi) { 
							if ($recipsbyline) { $MSMRTG::State::logdata{'mail'}[0]++;  }
							else {$MSMRTG::State::logdata{'mail'}[0] += $1; }
			}
     
			# The Mail logs (count messages)
      if ($mtainlog && /$mtamessre/oi) { $MSMRTG::State::logdata{'mail'}[1]++; }

      # The Spam logs
      elsif ($mtarejlog && /$mtaspamre/oi) { $blocked ++ }
    
      # Mailbytes if coming from MTA
      if ($mtainlog && $mbfrommta && /$mtabytesre/io)
        { $mailbytes += $1; }

    }
   
  }
  # Record the final position
  $MSMRTG::State::logfiles{$logfile}=tell(LOG);
  close LOG;

  # Ok. $blocked = the number of mails the MTA rejected with a 550
  # error code. This is the error code you should use to reject domains,
  # senders, receivers etc. you have problems with. We want the total
  # to be everything MailScanner & SpamAssassin caught ($spam) plus
  # everything rejected by the MTA with a 550 ($blocked) so the total
  # that shows up on mrtg will be all mails blocked, stopped, rejected
  # et. al. The mails that were blocked by the MTA with a 550 will
  # then show up as the red line on the graph.

  $MSMRTG::State::logdata{'spam'}[0] += ($spam + $blocked);
  $MSMRTG::State::logdata{'spam'}[1] += $blocked;

  # Similarly for viruses and problems, the total is all mails with problems
  # or viruses, the real viruses show up as the red line
  $MSMRTG::State::logdata{'virus'}[0] += ($viruses + $problems);
  $MSMRTG::State::logdata{'virus'}[1] += ($viruses);

	# And sort out the mailbytes figures
  $MSMRTG::State::logdata{'mailbytes'}[1]='u';

  # Mod mailbytes to convert in kB 
  $MSMRTG::State::logdata{'mailbytes'}[0]+=sprintf("%.0f",$mailbytes / (1024));

  if ($MSMRTG::Debug::debug) {
    MSMRTG::Debug::Debug("Leaving sub ProcessFile");
  }
  return;

}
sub GetRatios{
	# Get the spam and virus ratios
	# Takes 5 arguments, the previous values for 
	# mail, spam (total), spam (mta rules), virus (content checks), virus (scanner)
	
	my $mail=$_[0];
	my $spam1=$_[1];
	my $spam2=$_[2];
	my $virus1=$_[3];
	my $virus2=$_[4];

	my ($mailcount, $spamcount1, $spamcount2, $viruscount1, $viruscount2);
	
  if ($MSMRTG::Debug::debug) {
    MSMRTG::Debug::Debug("Entering sub GetRatios");
  }

	if ($MSMRTG::Config::Config{'ratiosaredaily'} =~ /yes/i) 
		{ $mail=$spam1=$spam2=$virus1=$virus2=0 };

	if ($mail eq 'u' || $spam1 eq 'u' || $spam2 eq 'u' || $virus1 eq 'u' || $virus2 eq 'u' || 
				$MSMRTG::State::logdata{'mail'}[0] eq 'u' ||
				$MSMRTG::State::logdata{'spam'}[0] eq 'u' ||
				$MSMRTG::State::logdata{'spam'}[1] eq 'u' ||
				$MSMRTG::State::logdata{'virus'}[0] eq 'u' ||
				$MSMRTG::State::logdata{'virus'}[1] eq 'u')
	{
		warn "ERROR: mail, spam and virus counters not fully initialised\n\tNo ratio data on this run\nThis is normal on the first run, but if you get this message every five minutes it may indicate problems reading the log files or writing the state file." ;
  	$MSMRTG::State::logdata{'spamratio'}[0]='u';
  	$MSMRTG::State::logdata{'spamratio'}[1]='u';
  	$MSMRTG::State::logdata{'virusratio'}[0]='u';
  	$MSMRTG::State::logdata{'virusratio'}[1]='u';
		return;
	}

	$mailcount=$MSMRTG::State::logdata{'mail'}[0]-$mail;

	# Skip if no mail processed 
	if ($mailcount==0){
  	$MSMRTG::State::logdata{'spamratio'}[0]='u';
	 	$MSMRTG::State::logdata{'spamratio'}[1]='u';
	 	$MSMRTG::State::logdata{'virusratio'}[0]='u';
	 	$MSMRTG::State::logdata{'virusratio'}[1]='u';

	  if ($MSMRTG::Debug::debug) {
 			MSMRTG::Debug::Debug("No mails processed - set ratios to 0");
 		}
		return;
	}
	
	# Check whether the user wants us to skip when only a few mails have
	# been processed...
	if (defined($MSMRTG::Config::Config{'skipratiosbelow'}))
		{
			if ($MSMRTG::Config::Config{'skipratiosbelow'}>$mailcount)
			{
  			$MSMRTG::State::logdata{'spamratio'}[0]='u';
		  	$MSMRTG::State::logdata{'spamratio'}[1]='u';
		  	$MSMRTG::State::logdata{'virusratio'}[0]='u';
		  	$MSMRTG::State::logdata{'virusratio'}[1]='u';

			  if ($MSMRTG::Debug::debug) {
    				MSMRTG::Debug::Debug("Too few mails, skipping ratios");
  			}
				return;
			}
		}
	
	$spamcount1=$MSMRTG::State::logdata{'spam'}[0]-$spam1;
	$spamcount2=$MSMRTG::State::logdata{'spam'}[1]-$spam2;
	# Notice we substract the number of actual viruses from this
	# to try and reduce duplication
	$viruscount1=($MSMRTG::State::logdata{'virus'}[0]-$MSMRTG::State::logdata{'virus'}[1])-$virus1;
	$viruscount2=$MSMRTG::State::logdata{'virus'}[1]-$virus2;

  $MSMRTG::State::logdata{'spamratio'}[0]
		=sprintf("%.0f",(($spamcount1*100)/$mailcount));
  $MSMRTG::State::logdata{'spamratio'}[1]
		=sprintf("%.0f",(($spamcount2*100)/$mailcount));
  $MSMRTG::State::logdata{'virusratio'}[0]
		=sprintf("%.0f",(($viruscount1*100)/$mailcount));
  $MSMRTG::State::logdata{'virusratio'}[1]
		=sprintf("%.0f",(($viruscount2*100)/$mailcount));

  if ($MSMRTG::Debug::debug) {
    MSMRTG::Debug::Debug("Leaving sub GetRatios");
  }
}
1;
